#include "main.h"

// 0..Pi/2 sine table
const uint8_t sine[64] PROGMEM = {
	0,     6,  13,  19,  25,  31,  37,  44,  50,  56,  62,  68,  74,  80,  86,  92,
	98,  103, 109, 115, 120, 126, 131, 136, 142, 147, 152, 157, 162, 167, 171, 176,
	180, 185, 189, 193, 197, 201, 205, 208, 212, 215, 219, 222, 225, 228, 231, 233,
	236, 238, 240, 242, 244, 246, 247, 249, 250, 251, 252, 253, 254, 254, 255, 255
};

// division bits
const uint8_t divs[16] PROGMEM = {4,4,4,4+8,4+8,4+8,2,2,2+8,2+4,2+4,2+4+8,2+4+8,1,1,1};

volatile register uint16_t acc     asm("r2");  // phase accumulator
volatile register uint16_t inc     asm("r4");  // phase increment
volatile register uint8_t  div     asm("r7");  // PWM amplitude division flags: 0=x1, 1=/2, 2=/4, 3=/8
volatile register uint8_t  tmr     asm("r8");  // event timer divisor
volatile register uint8_t  foff    asm("r9");  // on/off pin state
volatile register uint8_t  frev    asm("r10"); // reverse pin state
volatile register uint8_t  rev     asm("r6");  // current reverse state flag
volatile register uint8_t  adc     asm("r11"); // smoothed ADC
volatile register uint8_t  adc_inc asm("r12"); // ADC increment/decrement rate

int main(){
	// GPIO
	DDRB   = (1<<PB0)|(1<<PB1)|(1<<PB2)|(1<<PB4);
	// ADC
	ADMUX  = (1<<ADLAR)|(1<<MUX1)|(1<<MUX0);
	ADCSRA = (1<<ADEN)|(1<<ADSC)|(1<<ADATE)|(1<<ADPS2)|(1<<ADPS1)|(1<<ADPS0);
	ADCSRB = 0;
	// T0
	TCCR0A = (1<<WGM00);//|(1<<WGM01);
	TCCR0B = (1<<CS00);//(1<<CS01)
	TIMSK0 = (1<<TOIE0);
	sei();

	// PWM
	acc = inc = adc = 0;
	adc_inc = pgm_read_byte(ADDR_ADC_INC);
	if (adc_inc == 0xFF) adc_inc = ADC_INC;
	foff = PIN_IN_MAX;

	while(1){
		if (!tmr){ tmr = 255;

			if (foff >= (PIN_IN_MAX/2)){ // OFF state
				if (adc){
					if (adc >= adc_inc) adc -= adc_inc;
					else adc = 0;
				}
			} else { // ON state
				if (rev != (frev >= (PIN_IN_MAX/2))){ // reverse
					if (adc){                         // slow down to 0 RPM
						if (adc >= adc_inc) adc -= adc_inc;
						else adc = 0;
					} else rev = !rev;
				} else {                              // slow RPM up/down
					if (adc < ADCH){
						if (adc <= (ADC_MAX - adc_inc)) adc += adc_inc;
						else adc = ADC_MAX;
					} else if (adc > ADCH){
						if (adc >= adc_inc) adc -= adc_inc;
						else adc = 0;
					}
				}
				// get current amplitude divisor
				div = pgm_read_byte(&divs[adc >> 4]);
			}
			// calculate current RPM
			uint16_t rpm = (((uint16_t)adc * ((RPM_MAX-RPM_MIN) >> 4)) >> 4) + RPM_MIN;
			// calculate phase increment
			inc = (!adc) ? 0 : rpm >> 3;
		}
	}

	return 0;
}

ISR(TIM0_OVF_vect){
	if (tmr) tmr--;                                       // decrement event timer

	if (inc == 0){                                        // phase increment == 0 ? PWMs OFF
		TCCR0A &= ~((1<<COM0A1)|(1<<COM0B1));
		if (tmr == 0){
			PORTB |= (1<<PB2);
			DDRB  &= ~(1<<PB2);
			asm volatile ("nop \n nop \n");
			if (!!(PINB & (1<<PB2))){                     // read OFF pin
				if (foff <= (PIN_IN_MAX)) foff++;
			} else if (foff) foff--;
			DDRB  |= (1<<PB2);
		}
	} else {
		// PWMs ON
		TCCR0A |= (1<<COM0A1)|(1<<COM0B1);

		// PWM A
		uint8_t phase = acc >> 8;                         // phase MSB: 0..255
		if (phase & 0x80){                                // set bridge polarity
			PORTB |= (1<<PB4);
			DDRB  &= ~(1<<PB4);
			asm volatile ("nop \n nop \n");
			if (!!(PINB & (1<<PB4))){                    // read REV pin
				if (frev <= (PIN_IN_MAX)) frev++;
			} else if (frev) frev--;
			DDRB  |= (1<<PB4);
		} else PORTB &= ~(1<<PB4);
		uint8_t index = phase & 0x7F;                     // index in sine table
		if (index > 63) index = (uint8_t)127 - index;     // 90..180deg ? reverse index
		uint8_t ai = pgm_read_byte(&sine[index]), ao = 0; // undivided / divided amplitude
		if (div & 1) ao = ai; else { ai >>= 1;            // division
			if (div & 2) ao += ai; ai >>= 1;
			if (div & 4) ao += ai; ai >>= 1;
			if (div & 8) ao += ai;
		}
		OCR0A = ao;                                       // set PWM A regiter

		// PWM B
		phase += rev ? 192 : 64;                          // phase shifted by -/+90deg
		if (phase & 0x80){                                // set bridge polarity
			PORTB |= (1<<PB2);
			DDRB  &= ~(1<<PB2);
			asm volatile ("nop \n nop \n");
			if (!!(PINB & (1<<PB2))){                     // read OFF pin
				if (foff <= (PIN_IN_MAX)) foff++;
			} else if (foff) foff--;
			DDRB  |= (1<<PB2);
		} else PORTB &= ~(1<<PB2);
		index = phase & 0x7F;                             // index in sine table
		if (index > 63) index = (uint8_t)127 - index;     // 90..180deg ? reverse index
		ai = pgm_read_byte(&sine[index]), ao = 0;         // undivided / divided amplitude
		if (div & 1) ao = ai; else { ai >>= 1;            // division
			if (div & 2) ao += ai; ai >>= 1;
			if (div & 4) ao += ai; ai >>= 1;
			if (div & 8) ao += ai;
		}
		OCR0B = ao;                                       // set PWM B register

		acc += inc;                                       // next phase sample
	}
}
